package com.utopiaindustries.dao.ctp;

import com.utopiaindustries.model.ctp.FinishedItem;
import com.utopiaindustries.util.KeyHolderFunctions;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.jdbc.support.KeyHolder;
import org.springframework.stereotype.Repository;

import java.util.ArrayList;
import java.util.List;

@Repository
public class FinishedItemDAO {

	private final NamedParameterJdbcTemplate namedParameterJdbcTemplate;

	private final String TABLE_NAME = "cut_to_pack.finished_item";
	private final String SELECT_QUERY = String.format( "SELECT * FROM %s WHERE id = :id", TABLE_NAME );
	private final String SELECT_ALL_QUERY = String.format( "SELECT * FROM %s ORDER BY id DESC", TABLE_NAME );
	private final String DELETE_QUERY = String.format( "DELETE FROM %s WHERE id = :id", TABLE_NAME );
	private final String INSERT_QUERY = String.format( "INSERT INTO %s (id, item_id, sku, barcode, created_at, created_by, job_card_id, is_qa, stitched_item_id, is_segregated, qa_status) VALUES (:id, :item_id, :sku, :barcode, :created_at, :created_by, :job_card_id, :is_qa, :stitched_item_id, :is_segregated, :qa_status) ON DUPLICATE KEY UPDATE item_id = VALUES(item_id), sku = VALUES(sku), barcode = VALUES(barcode), created_at = VALUES(created_at), created_by = VALUES(created_by), job_card_id = VALUES(job_card_id), is_qa = VALUES(is_qa), stitched_item_id = VALUES(stitched_item_id), is_segregated = VALUES(is_segregated), qa_status = VALUES(qa_status)", TABLE_NAME );
	private final String SELECT_BY_LIMIT = String.format("SELECT * FROM %s ORDER BY id DESC LIMIT :limit", TABLE_NAME );
	private final String SELECT_BY_IDS = String.format( "SELECT * FROM %s WHERE id IN (:ids)", TABLE_NAME );
	private final String SELECT_BY_TERM = String.format( "SELECT * FROM %s WHERE barcode LIKE :term AND is_segregated = :is_segregated ORDER BY ID DESC", TABLE_NAME );
	private final String SELECT_BY_STITCHED_ITEM_ID = String.format( "SELECT * FROM %s WHERE stitched_item_id = :stitched_item_id", TABLE_NAME );
	private final String SELECT_BY_STITCHED_ITEM_IDS = String.format( "SELECT * FROM %s WHERE stitched_item_id IN (:stitched_item_ids)", TABLE_NAME );

	public FinishedItemDAO(NamedParameterJdbcTemplate namedParameterJdbcTemplate) {
		this.namedParameterJdbcTemplate = namedParameterJdbcTemplate;
	}

	// prepare query params
	private MapSqlParameterSource prepareInsertQueryParams( FinishedItem finishedItem ) {
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue( "id", finishedItem.getId() )
			.addValue( "item_id", finishedItem.getItemId() )
			.addValue( "sku", finishedItem.getSku() )
			.addValue( "barcode", finishedItem.getBarcode() )
			.addValue( "created_at", finishedItem.getCreatedAt() )
			.addValue( "created_by", finishedItem.getCreatedBy() )
			.addValue("job_card_id", finishedItem.getJobCardId() )
			.addValue("is_qa", finishedItem.getIsQa() )
			.addValue("stitched_item_id", finishedItem.getStitchedItemId() )
			.addValue("is_segregated", finishedItem.getIsSegregated() )
			.addValue("qa_status", finishedItem.getQaStatus() );
		return params;
	}

	// find
	public FinishedItem find( long id ) {
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue( "id", id );
		return  namedParameterJdbcTemplate.query( SELECT_QUERY, params, new FinishedItemRowMapper() )
				.stream()
				.findFirst()
				.orElse( new FinishedItem() );
	}

	// find all
	public List<FinishedItem> findAll() {
		return namedParameterJdbcTemplate.query( SELECT_ALL_QUERY, new FinishedItemRowMapper() );
	}

	// save
	public long save( FinishedItem finishedItem ) {
		KeyHolder keyHolder = new GeneratedKeyHolder();
		MapSqlParameterSource params = prepareInsertQueryParams( finishedItem );
		namedParameterJdbcTemplate.update( INSERT_QUERY, params, keyHolder );
		return KeyHolderFunctions.getKey( finishedItem.getId(), keyHolder );
	}

	// save all
	public int[] saveAll( List<FinishedItem> finishedItems ) {
		List<MapSqlParameterSource> batchArgs = new ArrayList<>();
		for ( FinishedItem finishedItem: finishedItems ) {
			MapSqlParameterSource params = prepareInsertQueryParams( finishedItem );
			batchArgs.add( params );
		}
		return namedParameterJdbcTemplate.batchUpdate( INSERT_QUERY, batchArgs.toArray(new MapSqlParameterSource[finishedItems.size()]) );
	}

	// delete
	public boolean delete( long id ) {
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue( "id", id );
		return namedParameterJdbcTemplate.update( DELETE_QUERY, params ) > 0;
	}

	public List<FinishedItem> findByLimit(Long count ){
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue( "limit",count );
		return namedParameterJdbcTemplate.query( SELECT_BY_LIMIT , params, new FinishedItemRowMapper() );
	}

	public List<FinishedItem> findByQuery(String query ){
		return namedParameterJdbcTemplate.query( query, new FinishedItemRowMapper() );
	}

	public List<FinishedItem> findByIds(List<Long> ids ){
		if( ids == null || ids.isEmpty() ) return new ArrayList<>();
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue("ids", ids );
		return namedParameterJdbcTemplate.query( SELECT_BY_IDS , params, new FinishedItemRowMapper() );
	}

	public List<FinishedItem> findByTerm( String term , boolean isSegregated ){
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue("term", "%" + term + "%" );
		params.addValue("is_segregated", isSegregated );
		return namedParameterJdbcTemplate.query( SELECT_BY_TERM , params, new FinishedItemRowMapper() );
	}


	public FinishedItem findByStitchedItem( long stitchedItemId ){
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue("stitched_item_id", stitchedItemId );
		return namedParameterJdbcTemplate.query( SELECT_BY_STITCHED_ITEM_ID , params, new FinishedItemRowMapper() )
				.stream()
				.findFirst()
				.orElse( null );

	}

	public List<FinishedItem> findByStitchedItemIds( List<Long> stitchedItemIds ){
		if( stitchedItemIds == null || stitchedItemIds.isEmpty() ) return new ArrayList<>();
		MapSqlParameterSource params = new MapSqlParameterSource();
		params.addValue("stitched_item_ids", stitchedItemIds );
		return namedParameterJdbcTemplate.query( SELECT_BY_STITCHED_ITEM_IDS, params, new FinishedItemRowMapper() );
	}
}