package com.utopiaindustries.service;

import com.google.zxing.BarcodeFormat;
import com.itextpdf.html2pdf.ConverterProperties;
import com.itextpdf.html2pdf.resolver.font.DefaultFontProvider;
import com.itextpdf.text.*;
import com.itextpdf.text.pdf.*;
import com.utopiaindustries.dao.ctp.BundleDAO;
import com.utopiaindustries.dao.ctp.FinishedItemDAO;
import com.utopiaindustries.dao.ctp.MasterBundleDAO;
import com.utopiaindustries.dao.ctp.StitchingOfflineItemDAO;
import com.utopiaindustries.model.ctp.*;
import com.utopiaindustries.util.BarcodeUtils;
import com.utopiaindustries.util.StringUtils;
import org.springframework.core.io.DefaultResourceLoader;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.ResourceLoader;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;

import java.util.ArrayList;
import java.util.List;

@Service
public class BarcodeService {

    private final BundleDAO bundleDAO;
    private final MasterBundleDAO masterBundleDAO;
    private final FinishedItemDAO finishedItemDAO;
    private final StitchingOfflineItemDAO stitchingOfflineItemDAO;

    public BarcodeService(BundleDAO bundleDAO, MasterBundleDAO masterBundleDAO, FinishedItemDAO finishedItemDAO, StitchingOfflineItemDAO stitchingOfflineItemDAO) {
        this.bundleDAO = bundleDAO;
        this.masterBundleDAO = masterBundleDAO;
        this.finishedItemDAO = finishedItemDAO;
        this.stitchingOfflineItemDAO = stitchingOfflineItemDAO;
    }

    /*
     *  generate barcodes here
     * */
    public ResponseEntity<InputStreamResource> generateBarcodes(List<Long> ids, String artifactType) throws Exception {
        String size = BarcodeStickerSize.SIZE_1_75_X_3_5.name();
        BarcodeStickerSize stickerSize = BarcodeStickerSize.getSize(size);

        List<? extends InventoryArtifact> list = new ArrayList<>();
        if (StringUtils.compare(artifactType, Bundle.class.getSimpleName())) {
            list = bundleDAO.findByIds(ids);
        } else if (StringUtils.compare(artifactType, MasterBundle.class.getSimpleName())) {
            list = masterBundleDAO.findByIds(ids);
        } else if (StringUtils.compare(artifactType, FinishedItem.class.getSimpleName())) {
            list = stitchingOfflineItemDAO.findByIds(ids);
        }

        // get input stream
        ByteArrayInputStream inputStream = null;
        // get input stream
        inputStream = getInputStream(list, stickerSize, artifactType);

        String headerContentDispositionStr = String.format("%s; filename=%s.pdf", "attachment", artifactType);
        // return response
        return ResponseEntity
                .ok()
                .header(HttpHeaders.CONTENT_DISPOSITION, headerContentDispositionStr)
                .contentType(MediaType.APPLICATION_PDF)
                .body(new InputStreamResource(inputStream));

    }

    private ByteArrayInputStream getInputStream(List<? extends InventoryArtifact> artifacts,
                                                BarcodeStickerSize stickerSize,
                                                String artifactType) throws Exception {
        // resource loader
        ResourceLoader resourceLoader = new DefaultResourceLoader();
        // output stream
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        // converter properties
        ConverterProperties properties = new ConverterProperties();
        properties.setFontProvider(new DefaultFontProvider(false, false, true));
        // generate pdf
        Document document = new Document(new Rectangle(stickerSize.getWidth(), stickerSize.getHeight()), stickerSize.getMarginLeft(), stickerSize.getMarginRight(), stickerSize.getMarginTop(), stickerSize.getMarginBottom());
        PdfWriter.getInstance(document, outputStream);
        document.open();
        for (InventoryArtifact artifact : artifacts) {
            document.newPage();
            // add sku
            Paragraph skuPara = new Paragraph(artifact.getSku(), new Font(Font.FontFamily.HELVETICA, stickerSize.getTextSizeSmall(), Font.BOLD));
            skuPara.setAlignment(Element.ALIGN_CENTER);
            document.add(skuPara);

            // add barcode image
            byte[] imgBytes = BarcodeUtils.getBarcodeImageByteArray(artifact.getBarcode(), BarcodeFormat.CODE_128, stickerSize.getImageWidthBarcode(), stickerSize.getImageHeightBarcode());
            Image image = Image.getInstance(imgBytes);
            image.setAlignment(Element.ALIGN_CENTER);
            document.add(image);

            Paragraph barcodePara = new Paragraph(artifact.getBarcode(), new Font(Font.FontFamily.HELVETICA, stickerSize.getTextSizeSmall(), Font.BOLD));
            barcodePara.setAlignment(Element.ALIGN_CENTER);
            document.add(barcodePara);
            // add Artifact mark B / M / F
            if ( artifactType.equalsIgnoreCase( Bundle.class.getSimpleName()) ) {
                Paragraph artifactPara = new Paragraph(String.format("%s : %d", artifact.getType(), artifact.getId() ), new Font(Font.FontFamily.HELVETICA, stickerSize.getTextSizeExtraSmall(), Font.NORMAL));
                artifactPara.setAlignment(Element.ALIGN_CENTER);

                Paragraph type = new Paragraph(String.valueOf(artifactType.toCharArray()[0]), new Font(Font.FontFamily.HELVETICA, stickerSize.getTextSize(), Font.BOLD));
                type.setAlignment(Element.ALIGN_CENTER);

                // Create a table with 2 columns
                PdfPTable table = new PdfPTable(2);
                table.setWidthPercentage(100);
                table.setWidths(new int[]{55, 45}); // Adjust column widths as needed

                // Add cells to the table
                PdfPCell artifactCell = new PdfPCell(artifactPara);
                artifactCell.setBorder(Rectangle.NO_BORDER);
                artifactCell.setHorizontalAlignment(Element.ALIGN_RIGHT);
                artifactCell.setPaddingTop(5);

                PdfPCell typeCell = new PdfPCell(type);
                typeCell.setBorder(Rectangle.NO_BORDER);
                typeCell.setHorizontalAlignment(Element.ALIGN_LEFT);

                table.addCell(artifactCell);
                table.addCell(typeCell);
                // Add the table to the document
                document.add(table);

            } else {
                Paragraph type = new Paragraph(String.valueOf(artifactType.toCharArray()[0]), new Font(Font.FontFamily.HELVETICA, stickerSize.getTextSize(), Font.BOLD));
                type.setAlignment(Element.ALIGN_CENTER);
                document.add(type);
            }
        }
        document.close();
        // input stream
        return new ByteArrayInputStream(outputStream.toByteArray());
    }


}